/*******************************************************************************
  MRF24WN Private Configuration Support

  File Name:
    wdrv_mrf24wn_iwpriv.c

  Summary:
    Configure optional (private) parameters of MRF24WN driver.

  Description:
    Functions in this module support the connection process for the
    MRF24W.
 *******************************************************************************/

//DOM-IGNORE-BEGIN
/*******************************************************************************
Copyright (c) 2013 released Microchip Technology Inc.  All rights reserved.

Microchip licenses to you the right to use, modify, copy and distribute
Software only when embedded on a Microchip microcontroller or digital signal
controller that is integrated into your product or third party product
(pursuant to the sublicense terms in the accompanying license agreement).

You should refer to the license agreement accompanying this Software for
additional information regarding your rights and obligations.

SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS WITHOUT WARRANTY OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
MERCHANTABILITY, TITLE, NON-INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE.
IN NO EVENT SHALL MICROCHIP OR ITS LICENSORS BE LIABLE OR OBLIGATED UNDER
CONTRACT, NEGLIGENCE, STRICT LIABILITY, CONTRIBUTION, BREACH OF WARRANTY, OR
OTHER LEGAL EQUITABLE THEORY ANY DIRECT OR INDIRECT DAMAGES OR EXPENSES
INCLUDING BUT NOT LIMITED TO ANY INCIDENTAL, SPECIAL, INDIRECT, PUNITIVE OR
CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF PROCUREMENT OF
SUBSTITUTE GOODS, TECHNOLOGY, SERVICES, OR ANY CLAIMS BY THIRD PARTIES
(INCLUDING BUT NOT LIMITED TO ANY DEFENSE THEREOF), OR OTHER SIMILAR COSTS.
 *******************************************************************************/

/******************/
/*    INCLUDES    */
/******************/
#include "wdrv_mrf24wn_iwpriv.h"
#include "wdrv_mrf24wn_scan_helper.h"
#include "wdrv_mrf24wn_priv.h"
#include "wdrv_mrf24wn_api.h"

/*******************/
/*    VARIABLES    */
/*******************/
extern WDRV_SCAN_RESULT_RAW g_scanResult;
extern WDRV_SCAN_RESULT g_scanResults[];

static uint8_t s_prescan = false;
static uint8_t s_prescan_waiting = false; // WF_PRESCAN - This is used only to allow prescan once.

/*******************/
/*    FUNCTIONS    */
/*******************/
void iwpriv_prescan_start(void)
{
    s_prescan_waiting = true;
    g_wdrv_priv.isScanDone = false;
    WDRV_PreScanStart();
}

bool iwpriv_prescan_done(void)
{
    if (s_prescan_waiting && g_wdrv_priv.isScanDone) {
        s_prescan_waiting = false;
        return true;
    }

    return false;
}

void iwpriv_prescan_set(bool scan)
{
    if (scan) {
        s_prescan = true;
        s_prescan_waiting = false;
    } else {
        s_prescan = false;
        s_prescan_waiting = false;
    }
}

bool iwpriv_prescan_get(void)
{
    return s_prescan;
}

void iwpriv_scan_start(void)
{
    WDRV_PreScanStart();
    SYS_CONSOLE_MESSAGE("Scan started ...\r\n");
    g_scanResult.displayIdx = 0;
    g_scanResult.getScanResultList = true;
}

IWPRIV_SCAN_RESULT iwpriv_scanstatus_get(void)
{
    IWPRIV_SCAN_RESULT ret;

    if(IS_SCAN_IN_PROGRESS(g_scanStatus.scanState)) {
        SYS_CONSOLE_MESSAGE((const char *)"Scan in progress.\r\n ");
        ret = IWPRIV_SCAN_IN_PROGRESS;
    } else if(IS_SCAN_STATE_DISPLAY(g_scanStatus.scanState) && (g_scanStatus.numScanResults > 0)){
        ret = IWPRIV_SCAN_SUCCEEDED;
    } else if (IS_SCAN_STATE_DISPLAY(g_scanStatus.scanState) && (g_scanStatus.numScanResults == 0)) {
        ret = IWPRIV_SCAN_NO_APS_FOUND;
    } else {
        ret = IWPRIV_SCAN_IDLE;
    }

    return ret;
}

uint16_t iwpriv_numberofscanresults_get(void)
{
    return g_scanStatus.numScanResults;
}

void *iwpriv_scanresult_get(uint8_t index)
{
    if (index <= iwpriv_numberofscanresults_get())
        return &g_scanResults[index];
    else
        return NULL;
}

IWPRIV_STATUS iwpriv_scanresults_display(void)
{
    if (IS_SCAN_STATE_DISPLAY(g_scanStatus.scanState)) {
        WDRV_ScanResultsDisplayManager();
        return IWPRIV_IN_PROGRESS;
    }

    return IWPRIV_READY;
}

IWPRIV_CONN_STATUS iwpriv_connstatus_get(void)
{
    IWPRIV_CONN_STATUS res = IWPRIV_CONNECTION_FAILED;
    WDRV_CONNECTION_STATES conn_status;

    conn_status = WDRV_ConnectStatus_Get();

    switch (conn_status) {
    case WDRV_CSTATE_NOT_CONNECTED:
        res = IWPRIV_CONNECTION_IDLE;
        break;

    case WDRV_CSTATE_CONNECTED_INFRASTRUCTURE:
        if (g_wdrv_priv.isConnReestablished) {
            g_wdrv_priv.isConnReestablished = false;
            res = IWPRIV_CONNECTION_REESTABLISHED;
            break;
        }
    case WDRV_CSTATE_CONNECTED_ADHOC:
        res = IWPRIV_CONNECTION_SUCCESSFUL;
        break;

    case WDRV_CSTATE_CONNECTION_IN_PROGRESS:
    case WDRV_CSTATE_RECONNECTION_IN_PROGRESS:
    case WDRV_CSTATE_CONNECTION_TEMPORARY_LOST:
        res = IWPRIV_CONNECTION_IN_PROGRESS;
        break;

    case WDRV_CSTATE_CONNECTION_PERMANENTLY_LOST:
        res = IWPRIV_CONNECTION_FAILED;
        break;

    default:
        WDRV_ASSERT(false, "Undefined connection status.");
    }

    return res;
}

void iwpriv_config_read(WDRV_CONFIG_DATA *wifi_config)
{
    memcpy(wifi_config, p_wdrv_configData, sizeof(WDRV_CONFIG_DATA));
}

void iwpriv_config_write(void *wifi_config)
{
    memcpy(p_wdrv_configData, wifi_config, sizeof(WDRV_CONFIG_DATA));
    WDRV_CONFIG_DataSave();
}

bool iwpriv_is_servermode(void)
{
    if (WDRV_CONFIG_PARAMS(networkType) == WDRV_NETWORK_TYPE_SOFT_AP || WDRV_CONFIG_PARAMS(networkType) == WDRV_NETWORK_TYPE_ADHOC)
        return true;
    else
        return false;
}

void iwpriv_devinfo_get(void *info)
{
    ((WDRV_DEVICE_INFO *)info)->deviceType = 3;
}

void iwpriv_ssid_set(uint8_t *ssid, uint8_t ssidLen)
{
    WDRV_EXT_CmdSSIDSet(ssid, ssidLen);
}

void iwpriv_ssid_get(uint8_t *ssid, uint8_t *ssidLen)
{
    memcpy(ssid, WDRV_CONFIG_PARAMS(netSSID), WDRV_CONFIG_PARAMS(SsidLength));
    ssid[WDRV_CONFIG_PARAMS(SsidLength)] = 0;
    *ssidLen = WDRV_CONFIG_PARAMS(SsidLength);
}

void iwpriv_nettype_set(uint8_t netType)
{
    if (netType == WDRV_NETWORK_TYPE_INFRASTRUCTURE) {
        WDRV_EXT_CmdNetModeBSSSet();
    } else if (netType == WDRV_NETWORK_TYPE_ADHOC) {
        WDRV_EXT_CmdNetModeIBSSSet();
    } else if (netType == WDRV_NETWORK_TYPE_SOFT_AP) {
        WDRV_EXT_CmdNetModeAPSet();
    }
}

void iwpriv_nettype_get(uint8_t *netType)
{
    *netType = WDRV_CONFIG_PARAMS(networkType);
}

void iwpriv_adhocctx_set(WDRV_ADHOC_NETWORK_CONTEXT *p_context)
{
    /* TODO: implementation is TBD */ 
}

IWPRIV_STATUS iwpriv_mcastfilter_set(uint8_t *addr)
{
    if (WDRV_MRF24WN_MulticastFilterSet(NULL, (TCPIP_MAC_ADDR *)addr) == TCPIP_MAC_RES_OK)
        return IWPRIV_READY;

    return IWPRIV_ERROR;
}

void iwpriv_powersave_config(bool enabled)
{    
    WDRV_EXT_CmdPowerSavePut(enabled); 
}

void iwpriv_initialconn_set(bool connect)
{
    g_wdrv_priv.initConn = connect;
}

void iwpriv_leftclient_get(bool *updated, TCPIP_MAC_ADDR *addr)
{
    /* TODO: Implement DHCP server sync function. */
    *updated = false;
}

uint8_t iwpriv_initstatus_get(void)
{
    if (g_wdrv_priv.initDriverDone)
        return IWPRIV_READY;
    else
        return IWPRIV_IN_PROGRESS;
}

//DOM-IGNORE-END
